from setuptools import setup, find_packages

setup(
    name='omnyx',
    version='0.0.1',
    description=(),
    long_description=(),
    author='tongbochen',
    author_email='bochen.tong@proton.me',
    url='',
    license='MIT Licence',
    packages=find_packages(),
    install_requires=[],
)
